using System;
using System.Collections.Generic;
using System.Text;
using System.Messaging;
using System.Xml.Serialization;
using NBus.Unicast.Transport;

namespace NBus.Unicast.Transport.Msmq
{
    //TODO: Support remote queue
    public class MsmqTransport : ITransport, IDisposable
    {
        #region config info

        public string InputQueue
        {
            set
            {
                MessageQueue q = new MessageQueue(value);

                if (!QueueIsLocal(value))
                    throw new ApplicationException("Cannot work transactionally with remote queue.");
                else
                    SetLocalQueue(q);

                this.queue.Formatter = new BinaryMessageFormatter();
            }
        }

        public string ErrorQueue
        {
            set
            {
                this.errorQueue = new MessageQueue(value);
            }
        }

        private bool useXmlSerialization;
        public bool UseXmlSerialization
        {
            set { useXmlSerialization = value; }
        }

        #endregion

        #region ITransport Members

        public string Address
        {
            get
            {
                return this.queue.Path;
            }
        }

        private bool isTransactional;
        public bool IsTransactional
        {
            set { this.isTransactional = value; }
        }

        private IList<Type> messageTypesToBeReceived;
        public IList<Type> MessageTypesToBeReceived
        {
            set
            {
                this.messageTypesToBeReceived = value;

                if (this.useXmlSerialization)
                    this.serializer = new XmlSerializer(typeof(object), this.GetExtraTypes(value));
            }
        }

        public void ClearAllMessages()
        {
            this.queue.Purge();
        }

        public void ReceiveMessageLater(Msg m)
        {
            //TODO: Update this to check for a remote queue
            this.Send(m, this.Address);
        }

        public void Send(Msg m, string destination)
        {
            string address = this.Resolve(destination);

            using (MessageQueue q = new MessageQueue(address, QueueAccessMode.Send))
            {
                q.Formatter = this.queue.Formatter;

                Message toSend = new Message();

                this.serializer.Serialize(toSend.BodyStream, new List<object>(m.Body));


                toSend.ResponseQueue = new MessageQueue(m.ReturnAddress);
                toSend.Label = m.WindowsIdentityName;

                if (m.CorrelationId != null)
                    toSend.CorrelationId = m.CorrelationId;

                foreach (IMessage message in m.Body)
                    if (message.GetType().GetCustomAttributes(typeof(RecoverableAttribute), true).Length > 0)
                        toSend.Recoverable = true;

                q.Send(toSend, this.GetTransactionTypeForSend());

                m.Id = toSend.Id;
            }
        }

        public Msg Receive(TimeSpan timeout)
        {
            try
            {
                Message m = this.queue.Receive(timeout, this.GetTransactionTypeForReceive());

                Msg result = Convert(m);

                try
                {
                    result.Body = this.Extract(m);
                }
                catch
                {
                    if (this.errorQueue != null)
                        this.errorQueue.Send(m, this.GetTransactionTypeForSend());

                    return null;
                }

                return result;

            }
            catch (MessageQueueException mqe)
            {
                if (mqe.MessageQueueErrorCode == MessageQueueErrorCode.IOTimeout)
                    return null;
            }

            return null;
        }

        #endregion

        #region helper methods

        public static bool QueueIsLocal(string value)
        {
            string machineName = Environment.MachineName.ToLower();

            value = value.Replace("FormatName:DIRECT=OS:", "");
            int index = value.IndexOf('\\');

            string queueMachineName = value.Substring(0, index).ToLower();

            return (machineName == queueMachineName || queueMachineName == "localhost" || queueMachineName == ".");
        }

        public static Msg Convert(Message m)
        {
            Msg result = new Msg();
            result.Id = m.Id;
            result.CorrelationId = m.CorrelationId;

            if (m.ResponseQueue != null)
                result.ReturnAddress = m.ResponseQueue.Path;

            result.WindowsIdentityName = m.Label;

            return result;
        }

        private string Resolve(string destination)
        {
            string dest = destination.ToUpper().Replace("FORMATNAME:DIRECT=OS:", "");

            MessageQueue q = new MessageQueue(dest);
            if (QueueIsLocal(dest))
                q.MachineName = Environment.MachineName;

            return "FormatName:DIRECT=OS:" + q.Path;
        }

        private IMessage[] Extract(Message message)
        {
            IMessage[] result;

            if (this.serializer != null)
            {
                List<object> body = serializer.Deserialize(message.BodyStream) as List<object>;

                result = new IMessage[body.Count];

                int i = 0;
                foreach (IMessage m in body)
                    result[i++] = m;
            }
            else
                result = message.Body as IMessage[];

            return result;
        }

        private MessageQueueTransactionType GetTransactionTypeForReceive()
        {
            if (this.isTransactional)
                return MessageQueueTransactionType.Automatic;
            else
                return MessageQueueTransactionType.None;
        }

        private MessageQueueTransactionType GetTransactionTypeForSend()
        {
            if (this.isTransactional)
                return MessageQueueTransactionType.Automatic;
            else
                return MessageQueueTransactionType.Single;
        }

        private void SetLocalQueue(MessageQueue q)
        {
            q.MachineName = Environment.MachineName; // just in case we were given "localhost"

            if (!q.Transactional)
                throw new ArgumentException("Queue must be transactional (" + q.Path + ").");
            else
                this.queue = q;

            MessagePropertyFilter mpf = new MessagePropertyFilter();
            mpf.SetAll();

            this.queue.MessageReadPropertyFilter = mpf;
        }

        private Type[] GetExtraTypes(IList<Type> value)
        {
            value.Add(typeof(List<object>));

            List<Type> list = value as List<Type>;

            if (list != null)
                return list.ToArray();

            return new List<Type>(value).ToArray();
        }

        #endregion

        #region members

        private MessageQueue queue;
        private MessageQueue errorQueue;
        private XmlSerializer serializer = null;

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            this.queue.Dispose();
        }

        #endregion
    }
}
